import { transform, last, isUndefined, keys, toString } from 'micro-dash';

/**
 * Returns whether `set` contains all the elements in `subset`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4]);
 * setUnion(setA, setB); // true
 * ```
 */
function isSuperset(set, subset) {
    for (const elem of subset) {
        if (!set.has(elem)) {
            return false;
        }
    }
    return true;
}

function isSetEqual(setA, setB) {
    return setA.size === setB.size && isSuperset(setA, setB);
}

/**
 * Returns a new set that contains the elements of `setA` that are not in `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setDifference(setA, setB); // Set [1, 2]
 * ```
 */
function setDifference(setA, setB) {
    const difference = new Set(setA);
    for (const elem of setB) {
        difference.delete(elem);
    }
    return difference;
}

/**
 * Returns a new set that contains the elements that are in both `setA` and `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setIntersection(setA, setB); // Set [3, 4]
 * ```
 */
function setIntersection(setA, setB) {
    const intersection = new Set();
    for (const elem of setB) {
        if (setA.has(elem)) {
            intersection.add(elem);
        }
    }
    return intersection;
}

/**
 * Returns a new set that contains the elements that are in either `setA` or `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setUnion(setA, setB); // Set [1, 2, 3, 4, 5, 6]
 * ```
 */
function setUnion(setA, setB) {
    const union = new Set(setA);
    for (const elem of setB) {
        if (!setA.has(elem)) {
            union.add(elem);
        }
    }
    return union;
}

/**
 * Returns a new set that contains the elements that are in only one of the two sets. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * symmetricSetDifference(setA, setB); // Set [1, 2, 5, 6]
 * ```
 */
function symmetricSetDifference(setA, setB) {
    const difference = new Set(setA);
    for (const elem of setB) {
        if (difference.has(elem)) {
            difference.delete(elem);
        }
        else {
            difference.add(elem);
        }
    }
    return difference;
}

/**
 * Like a standard debounce from e.g. `micro-dash`, but can execute a different function for a different wait period at each invocation.
 *
 * ```ts
 * const print = (value) => console.log(value);
 * const debouncer = new Debouncer();
 *
 * debouncer.run(print, 1000, 1);
 * await sleep(500);
 *
 * debouncer.run(print, 1000, 2);
 * await sleep(1000); // prints "2"
 *
 * debouncer.run(print, 0, 3);
 * debouncer.run(print, 1000, 4);
 * await sleep(500);
 *
 * debouncer.run(print, 2000, 5);
 * debouncer.run(print, 50, 6);
 * await sleep(50); // prints "6"
 * ```
 */
class Debouncer {
    run(func, wait = 0, ...args) {
        this.cancel();
        this.timeoutId = setTimeout(func, wait, ...args);
    }
    cancel() {
        if (this.timeoutId !== undefined) {
            clearTimeout(this.timeoutId);
        }
    }
}

/**
 * Returns a new function to use in place of `func` that will call the provided hooks in addition to `func`. They are called in the following order:
 *
 * 1. `before`
 * 2. `around` if provided, else `original`
 * 3. `transform`
 * 4. `after`
 *
 * ```ts
 * const sum = (a: number, b: number) => a + b;
 * const sumAndLog = wrapFunction(sum, {
 *   after: (result, a, b) => {
 *     console.log(a, '+', b, '=', result);
 *   },
 * }
 * const sumPlusOne = wrapFunction(sum, {
 *   transform: (result) => result + 1,
 * });
 * ```
 */
function wrapFunction(original, hooks) {
    const wrapped = function (...args) {
        let result;
        callHook(hooks.before, this, args);
        if (hooks.around) {
            result = hooks.around.call(this, original, ...args);
        }
        else {
            result = original.apply(this, args);
        }
        result = callHook(hooks.transform, this, [result, ...args], result);
        callHook(hooks.after, this, [result, ...args]);
        return result;
    };
    Object.defineProperty(wrapped, "length", { value: original.length });
    return wrapped;
}
/** @hidden */
function callHook(hook, context, args, defaultResult) {
    return hook ? hook.apply(context, args) : defaultResult;
}

/**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 */
class Deferred {
    constructor() {
        this.pending = true;
        const hooks = {
            before: () => {
                this.pending = false;
            },
        };
        this.promise = new Promise((resolve, reject) => {
            this.resolve = wrapFunction(resolve, hooks);
            this.reject = wrapFunction(reject, hooks);
        });
    }
    isPending() {
        return this.pending;
    }
}

/**
 * Returns a promise that resolves after `ms` milliseconds.
 *
 * ```ts
 * // do something
 * await sleep(1000); // wait a second
 * // do something else
 * ```
 */
function sleep(ms) {
    return new Promise((resolve) => {
        setTimeout(resolve, ms);
    });
}

function mapToObject(collection, iteratee) {
    return transform(collection, (accumulator, origValue, keyOrIndex) => {
        const [key, newValue] = iteratee(origValue, keyOrIndex);
        accumulator[key] = newValue;
    });
}

function mapAsKeys(collection, iteratee) {
    return mapToObject(collection, (value, keyOrIndex) => [
        value,
        iteratee(value, keyOrIndex),
    ]);
}

/**
 * Rounds `value` to the nearest multiple of `multiple`.
 *
 * ```ts
 * roundToMultiple(5, 13); // 15
 * roundToMultiple(2, 4.8); // 4
 * roundToMultiple(3, -4); // -3
 * roundToMultiple(0.5, 1.6); // 1.5
 * ```
 */
function roundToMultipleOf(multiple, value) {
    return Math.round(value / multiple) * multiple;
}

/**
 * Defines the canonical string representation for each time unit. Many aliases
 * can also be used with the functions that deal with time units:
 *
 * - ns, Nanosecond, Nanoseconds, nanosecond, nanoseconds, nanos
 * - μs, Microsecond, Microseconds, microsecond, microseconds, micros
 * - ms, Millisecond, Milliseconds, millisecond, milliseconds, millis
 * - s, Second, Seconds, second, seconds, S, sec, sec, secs
 * - m, Minute, Minutes, minute, minutes, M, min, mins
 * - h, Hour, Hours, hour, hours, H, hr, hrs
 * - d, Day, Days, day, days, D
 * - w, Week, Weeks, week, weeks, W, wk, wks
 * - y, Year, Years, year, years, Y, yr, yrs
 * - dec, Decade, Decades, decade, decades
 * - cent, Century, Centuries, century, centuries
 * - mil, Millennium, Millennia, millennium, millennia
 */
var TimeUnit;
(function (TimeUnit) {
    TimeUnit["Nanoseconds"] = "ns";
    TimeUnit["Microseconds"] = "\u03BCs";
    TimeUnit["Milliseconds"] = "ms";
    TimeUnit["Seconds"] = "s";
    TimeUnit["Minutes"] = "m";
    TimeUnit["Hours"] = "h";
    TimeUnit["Days"] = "d";
    TimeUnit["Weeks"] = "w";
    TimeUnit["Years"] = "y";
    TimeUnit["Decades"] = "dec";
    TimeUnit["Centuries"] = "cent";
    TimeUnit["Millennia"] = "mil";
})(TimeUnit || (TimeUnit = {}));
/** @hidden */
const nanoConversions = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, getNanoConversions(TimeUnit.Nanoseconds, "Nanosecond", 1, {
    aliases: ["nanos"],
})), getNanoConversions(TimeUnit.Microseconds, "Microsecond", 1000, {
    aliases: ["micros"],
})), getNanoConversions(TimeUnit.Milliseconds, "Millisecond", 1000 * 1000, {
    aliases: ["millis"],
})), getNanoConversions(TimeUnit.Seconds, "Second", 1000 * 1000 * 1000, {
    aliases: ["S", "sec", "secs"],
})), getNanoConversions(TimeUnit.Minutes, "Minute", 60 * 1000 * 1000 * 1000, {
    aliases: ["M", "min", "mins"],
})), getNanoConversions(TimeUnit.Hours, "Hour", 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["H", "hr", "hrs"],
})), getNanoConversions(TimeUnit.Days, "Day", 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["D"],
})), getNanoConversions(TimeUnit.Weeks, "Week", 7 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["W", "wk", "wks"],
})), getNanoConversions(TimeUnit.Years, "Year", 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["Y", "yr", "yrs"],
})), getNanoConversions(TimeUnit.Decades, "Decade", 10 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000)), getNanoConversions(TimeUnit.Centuries, "Century", 100 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Centuries" })), getNanoConversions(TimeUnit.Millennia, "Millennium", 1000 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Millennia" }));
/**
 * Converts time between two units. Units can be any value described in the docs for [[TimeUnit]].
 *
 * ```ts
 * convertTime(1, 's', 'ms'); // 1000
 * convertTime(20, TimeUnit.Decades, TimeUnit.Centuries); // 2
 * ```
 */
function convertTime(value, unit, targetUnit) {
    return (value * nanoConversions[unit]) / nanoConversions[targetUnit];
}
/**
 * Constructs a string representation of an elapsed amount of time. The least significant unit will be rounded to the nearest whole number.
 *
 * ```ts
 * elapsedToString(2001, ["s", "ms"]); // "2 s 1 ms"
 * elapsedToString(15, ["wks", "d"], ( elapsedUnit: TimeUnit.Days }); // "2 wks 1 d"
 * elapsedToString(1, [TimeUnit.Microseconds]); // "1000 μs"
 * ```
 *
 * @param showLeadingZeros whether to include the most significant units in the string if they are zero.
 */
function elapsedToString(elapsed, units, { elapsedUnit = TimeUnit.Milliseconds, showLeadingZeros = true } = {}) {
    elapsed = roundToMultipleOf(convertTime(1, last(units), elapsedUnit), elapsed);
    let showZeros = showLeadingZeros;
    const tokens = [];
    units.forEach((unit, i) => {
        const conversion = convertTime(1, elapsedUnit, unit);
        const value = Math.floor(elapsed * conversion);
        if (value > 0 || showZeros || i === units.length - 1) {
            tokens.push(value, unit);
            showZeros = true;
        }
        elapsed -= value / conversion;
    });
    return tokens.join(" ");
}
/** @hidden */
function getNanoConversions(unit, singular, nanos, { aliases = [], plural = singular + "s" } = {}) {
    return mapAsKeys([
        unit,
        singular,
        singular.toLowerCase(),
        plural,
        plural.toLowerCase(),
        ...aliases,
    ], () => nanos);
}

/**
 * Throws an error if `condition` is falsy, and acts as a type guard.
 *
 * @param message The message to set in the error that is thrown
 *
 * ```ts
 * class ProblemLogger {
 *   problem?: string;
 *
 *   log() {
 *     assert(this.problem, "You must set problem before logging it");
 *
 *     // now typescript knows `this.problem` is truthy, so we can safely call methods on it
 *     console.log(problem.toUpperCase());
 *   }
 * }
 * ```
 */
function assert(condition, message) {
    if (!condition) {
        throw new Error(message);
    }
}

/**
 * Extend this for classes whose objects are directly callable.
 *
 * ```ts
 * class Multiplier extends CallableObject<(value: number) => number> {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 */
class CallableObject {
    constructor(callAction) {
        return Object.setPrototypeOf(callAction, new.target.prototype);
    }
}

/** @hidden */
let seq = 0;
/**
 * Creates a factory function for building objects of a given type. Commonly used to create test objects for use in specs.
 *
 * ```ts
 * interface Message {
 *   id: number;
 *   text: string;
 * }
 *
 * interface Options {
 *   cypherDistance: number;
 * }
 *
 * declare function shiftCharacters(text: string, distance: number): string;
 *
 * const buildMessage = createBuilder<Message, Options>(
 *   (seq) => ({ id: seq, text: `message ${seq}` }),
 *   (message, _seq, options) => {
 *     message.text = shiftCharacters(message.text, options.cypherDistance || 0);
 *   },
 * );
 *
 * buildMessage(); // { id: 1, text: "message 1" }
 * buildMessage({ text: "Hello world!" }); // { id: 2, text: "Hello world!" }
 * buildMessage({ text: "abc" }, { cypherDistance: 3 }); // { id: 3, text: "def" }
 * ```
 */
function createBuilder(buildDefaults, afterBuild) {
    return (attributes, options = {}) => {
        ++seq;
        const obj = Object.assign(buildDefaults(seq, options), attributes);
        if (afterBuild) {
            afterBuild(obj, seq, options);
        }
        return obj;
    };
}

/**
 * Checks if `value` is anything other than `undefined`.
 *
 * ```ts
 * isDefined({ a: 1 }); // true
 * isDefined(false); // true
 * isDefined(null); // true
 * isDefined(undefined); // false
 * ```
 */
function isDefined(value) {
    return !isUndefined(value);
}

/**
 * Performs a deep comparison between two values to determine if they are equivalent, up to the given depth. Once that depth is reached, remaining values are compared using `Object.is()`.
 *
 * ```ts
 * let object1 = "a";
 * let object2 = "a";
 * isEqualAtDepth(0, object1, object2); // true
 * isEqualAtDepth(1, object1, object2); // true
 *
 * let object1 = { a: 1 };
 * let object2 = { a: 1 };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // true
 *
 * object1 = { a: [1, 2, 3], d: { e: 1 } };
 * object2 = { a: [1, 2, 3], d: { e: 1 } };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // false
 * isEqualAtDepth(2, object1, object2); // true
 * isEqualAtDepth(3, object1, object2); // true
 * ```
 */
function isEqualAtDepth(depth, value, other) {
    if (Object.is(value, other)) {
        // covers e.g. NaN === NaN
        return true;
    }
    if (depth === 0 || !(value instanceof Object && other instanceof Object)) {
        return false;
    }
    return hasSameValues(depth, value, other);
}
/** @hidden */
function hasSameValues(depth, value, other) {
    for (const key of keys(value)) {
        if (!other.hasOwnProperty(key)) {
            return false;
        }
    }
    for (const key of keys(other)) {
        if (!isEqualAtDepth(depth - 1, value[key], other[key])) {
            return false;
        }
    }
    return true;
}

/**
 * Converts a 2D array to a csv string. Values are converted using micro-dash's toString().
 *
 * ```ts
 * toCsv([["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"]]); // "a,b,c\nd,e,f\ng,h,i"
 * toCsv([
 *   ["a", "", "string"]
 *   [undefined, null],
 *   [true, false],
 *   [1, 2, 3],
 *   [{}, { hi: "there" }]
 * ]) // "a,,string\n,\ntrue,false\n1,2,3\n[object Object],[object Object]"
 * ```
 */
function toCsv(content) {
    return content
        .map((row) => row.map((cell) => toCellString(cell)).join(","))
        .join("\n");
}
/** @hidden */
const specialCsvCharactersRegexp = /["|,|\n|\r]/;
/** @hidden */
const allDoubleQuotes = /"/g;
/** @hidden */
function toCellString(value) {
    const str = toString(value);
    if (specialCsvCharactersRegexp.test(str)) {
        return `"${str.replace(allDoubleQuotes, `""`)}"`;
    }
    else {
        return str;
    }
}

/**
 * Replaces a method on `object` with a wrapped version that will call the provided hooks in addition to the original method. See `wrapFunction()` for more details on the hooks.
 *
 * @returns a function to reset the method to its previous, unwrapped state
 *
 * ```ts
 * // log all get requests to the console
 * wrapMethod(HttpClient.prototype, "get", {
 *   before(url) {
 *     console.log("Sending GET request to", url);
 *   }
 * });
 *
 * // suppress benign error messages
 * const unwrap = wrapMethod(console, "error", {
 *   around(original, ...args) {
 *     if (args[0].message !== 'something benign') {
 *       original(...args);
 *     }
 *   }
 * });
 *
 * // remove error suppression (from above)
 * unwrap();
 * ```
 */
function wrapMethod(object, key, hooks) {
    const original = object[key];
    object[key] = wrapFunction(original, hooks);
    return () => {
        object[key] = original;
    };
}

/*
 * Public API Surface of s-js-utils
 */

/**
 * Generated bundle index. Do not edit.
 */

export { CallableObject, Debouncer, Deferred, TimeUnit, assert, convertTime, createBuilder, elapsedToString, isDefined, isEqualAtDepth, isSetEqual, isSuperset, mapAsKeys, mapToObject, roundToMultipleOf, setDifference, setIntersection, setUnion, sleep, symmetricSetDifference, toCsv, wrapFunction, wrapMethod };
//# sourceMappingURL=s-js-utils.js.map
