import { __values, __spread, __read, __assign } from 'tslib';
import { transform, last, isUndefined, keys, toString } from 'micro-dash';

/**
 * Returns whether `set` contains all the elements in `subset`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4]);
 * setUnion(setA, setB); // true
 * ```
 */
function isSuperset(set, subset) {
    var e_1, _a;
    try {
        for (var subset_1 = __values(subset), subset_1_1 = subset_1.next(); !subset_1_1.done; subset_1_1 = subset_1.next()) {
            var elem = subset_1_1.value;
            if (!set.has(elem)) {
                return false;
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (subset_1_1 && !subset_1_1.done && (_a = subset_1.return)) _a.call(subset_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return true;
}

function isSetEqual(setA, setB) {
    return setA.size === setB.size && isSuperset(setA, setB);
}

/**
 * Returns a new set that contains the elements of `setA` that are not in `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setDifference(setA, setB); // Set [1, 2]
 * ```
 */
function setDifference(setA, setB) {
    var e_1, _a;
    var difference = new Set(setA);
    try {
        for (var setB_1 = __values(setB), setB_1_1 = setB_1.next(); !setB_1_1.done; setB_1_1 = setB_1.next()) {
            var elem = setB_1_1.value;
            difference.delete(elem);
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (setB_1_1 && !setB_1_1.done && (_a = setB_1.return)) _a.call(setB_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return difference;
}

/**
 * Returns a new set that contains the elements that are in both `setA` and `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setIntersection(setA, setB); // Set [3, 4]
 * ```
 */
function setIntersection(setA, setB) {
    var e_1, _a;
    var intersection = new Set();
    try {
        for (var setB_1 = __values(setB), setB_1_1 = setB_1.next(); !setB_1_1.done; setB_1_1 = setB_1.next()) {
            var elem = setB_1_1.value;
            if (setA.has(elem)) {
                intersection.add(elem);
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (setB_1_1 && !setB_1_1.done && (_a = setB_1.return)) _a.call(setB_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return intersection;
}

/**
 * Returns a new set that contains the elements that are in either `setA` or `setB`. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * setUnion(setA, setB); // Set [1, 2, 3, 4, 5, 6]
 * ```
 */
function setUnion(setA, setB) {
    var e_1, _a;
    var union = new Set(setA);
    try {
        for (var setB_1 = __values(setB), setB_1_1 = setB_1.next(); !setB_1_1.done; setB_1_1 = setB_1.next()) {
            var elem = setB_1_1.value;
            if (!setA.has(elem)) {
                union.add(elem);
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (setB_1_1 && !setB_1_1.done && (_a = setB_1.return)) _a.call(setB_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return union;
}

/**
 * Returns a new set that contains the elements that are in only one of the two sets. Taken from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Set.
 *
 * ```ts
 * const setA = new Set([1, 2, 3, 4]);
 * const setB = new Set([3, 4, 5, 6]);
 * symmetricSetDifference(setA, setB); // Set [1, 2, 5, 6]
 * ```
 */
function symmetricSetDifference(setA, setB) {
    var e_1, _a;
    var difference = new Set(setA);
    try {
        for (var setB_1 = __values(setB), setB_1_1 = setB_1.next(); !setB_1_1.done; setB_1_1 = setB_1.next()) {
            var elem = setB_1_1.value;
            if (difference.has(elem)) {
                difference.delete(elem);
            }
            else {
                difference.add(elem);
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (setB_1_1 && !setB_1_1.done && (_a = setB_1.return)) _a.call(setB_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return difference;
}

/**
 * Like a standard debounce from e.g. `micro-dash`, but can execute a different function for a different wait period at each invocation.
 *
 * ```ts
 * const print = (value) => console.log(value);
 * const debouncer = new Debouncer();
 *
 * debouncer.run(print, 1000, 1);
 * await sleep(500);
 *
 * debouncer.run(print, 1000, 2);
 * await sleep(1000); // prints "2"
 *
 * debouncer.run(print, 0, 3);
 * debouncer.run(print, 1000, 4);
 * await sleep(500);
 *
 * debouncer.run(print, 2000, 5);
 * debouncer.run(print, 50, 6);
 * await sleep(50); // prints "6"
 * ```
 */
var Debouncer = /** @class */ (function () {
    function Debouncer() {
    }
    Debouncer.prototype.run = function (func, wait) {
        if (wait === void 0) { wait = 0; }
        var args = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            args[_i - 2] = arguments[_i];
        }
        this.cancel();
        this.timeoutId = setTimeout.apply(void 0, __spread([func, wait], args));
    };
    Debouncer.prototype.cancel = function () {
        if (this.timeoutId !== undefined) {
            clearTimeout(this.timeoutId);
        }
    };
    return Debouncer;
}());

/**
 * Returns a new function to use in place of `func` that will call the provided hooks in addition to `func`. They are called in the following order:
 *
 * 1. `before`
 * 2. `around` if provided, else `original`
 * 3. `transform`
 * 4. `after`
 *
 * ```ts
 * const sum = (a: number, b: number) => a + b;
 * const sumAndLog = wrapFunction(sum, {
 *   after: (result, a, b) => {
 *     console.log(a, '+', b, '=', result);
 *   },
 * }
 * const sumPlusOne = wrapFunction(sum, {
 *   transform: (result) => result + 1,
 * });
 * ```
 */
function wrapFunction(original, hooks) {
    var wrapped = function () {
        var _a;
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var result;
        callHook(hooks.before, this, args);
        if (hooks.around) {
            result = (_a = hooks.around).call.apply(_a, __spread([this, original], args));
        }
        else {
            result = original.apply(this, args);
        }
        result = callHook(hooks.transform, this, __spread([result], args), result);
        callHook(hooks.after, this, __spread([result], args));
        return result;
    };
    Object.defineProperty(wrapped, "length", { value: original.length });
    return wrapped;
}
/** @hidden */
function callHook(hook, context, args, defaultResult) {
    return hook ? hook.apply(context, args) : defaultResult;
}

/**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 */
var Deferred = /** @class */ (function () {
    function Deferred() {
        var _this = this;
        this.pending = true;
        var hooks = {
            before: function () {
                _this.pending = false;
            },
        };
        this.promise = new Promise(function (resolve, reject) {
            _this.resolve = wrapFunction(resolve, hooks);
            _this.reject = wrapFunction(reject, hooks);
        });
    }
    Deferred.prototype.isPending = function () {
        return this.pending;
    };
    return Deferred;
}());

/**
 * Returns a promise that resolves after `ms` milliseconds.
 *
 * ```ts
 * // do something
 * await sleep(1000); // wait a second
 * // do something else
 * ```
 */
function sleep(ms) {
    return new Promise(function (resolve) {
        setTimeout(resolve, ms);
    });
}

function mapToObject(collection, iteratee) {
    return transform(collection, function (accumulator, origValue, keyOrIndex) {
        var _a = __read(iteratee(origValue, keyOrIndex), 2), key = _a[0], newValue = _a[1];
        accumulator[key] = newValue;
    });
}

function mapAsKeys(collection, iteratee) {
    return mapToObject(collection, function (value, keyOrIndex) { return [
        value,
        iteratee(value, keyOrIndex),
    ]; });
}

/**
 * Rounds `value` to the nearest multiple of `multiple`.
 *
 * ```ts
 * roundToMultiple(5, 13); // 15
 * roundToMultiple(2, 4.8); // 4
 * roundToMultiple(3, -4); // -3
 * roundToMultiple(0.5, 1.6); // 1.5
 * ```
 */
function roundToMultipleOf(multiple, value) {
    return Math.round(value / multiple) * multiple;
}

/**
 * Defines the canonical string representation for each time unit. Many aliases
 * can also be used with the functions that deal with time units:
 *
 * - ns, Nanosecond, Nanoseconds, nanosecond, nanoseconds, nanos
 * - μs, Microsecond, Microseconds, microsecond, microseconds, micros
 * - ms, Millisecond, Milliseconds, millisecond, milliseconds, millis
 * - s, Second, Seconds, second, seconds, S, sec, sec, secs
 * - m, Minute, Minutes, minute, minutes, M, min, mins
 * - h, Hour, Hours, hour, hours, H, hr, hrs
 * - d, Day, Days, day, days, D
 * - w, Week, Weeks, week, weeks, W, wk, wks
 * - y, Year, Years, year, years, Y, yr, yrs
 * - dec, Decade, Decades, decade, decades
 * - cent, Century, Centuries, century, centuries
 * - mil, Millennium, Millennia, millennium, millennia
 */
var TimeUnit;
(function (TimeUnit) {
    TimeUnit["Nanoseconds"] = "ns";
    TimeUnit["Microseconds"] = "\u03BCs";
    TimeUnit["Milliseconds"] = "ms";
    TimeUnit["Seconds"] = "s";
    TimeUnit["Minutes"] = "m";
    TimeUnit["Hours"] = "h";
    TimeUnit["Days"] = "d";
    TimeUnit["Weeks"] = "w";
    TimeUnit["Years"] = "y";
    TimeUnit["Decades"] = "dec";
    TimeUnit["Centuries"] = "cent";
    TimeUnit["Millennia"] = "mil";
})(TimeUnit || (TimeUnit = {}));
/** @hidden */
var nanoConversions = __assign(__assign(__assign(__assign(__assign(__assign(__assign(__assign(__assign(__assign(__assign(__assign({}, getNanoConversions(TimeUnit.Nanoseconds, "Nanosecond", 1, {
    aliases: ["nanos"],
})), getNanoConversions(TimeUnit.Microseconds, "Microsecond", 1000, {
    aliases: ["micros"],
})), getNanoConversions(TimeUnit.Milliseconds, "Millisecond", 1000 * 1000, {
    aliases: ["millis"],
})), getNanoConversions(TimeUnit.Seconds, "Second", 1000 * 1000 * 1000, {
    aliases: ["S", "sec", "secs"],
})), getNanoConversions(TimeUnit.Minutes, "Minute", 60 * 1000 * 1000 * 1000, {
    aliases: ["M", "min", "mins"],
})), getNanoConversions(TimeUnit.Hours, "Hour", 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["H", "hr", "hrs"],
})), getNanoConversions(TimeUnit.Days, "Day", 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["D"],
})), getNanoConversions(TimeUnit.Weeks, "Week", 7 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["W", "wk", "wks"],
})), getNanoConversions(TimeUnit.Years, "Year", 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["Y", "yr", "yrs"],
})), getNanoConversions(TimeUnit.Decades, "Decade", 10 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000)), getNanoConversions(TimeUnit.Centuries, "Century", 100 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Centuries" })), getNanoConversions(TimeUnit.Millennia, "Millennium", 1000 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Millennia" }));
/**
 * Converts time between two units. Units can be any value described in the docs for [[TimeUnit]].
 *
 * ```ts
 * convertTime(1, 's', 'ms'); // 1000
 * convertTime(20, TimeUnit.Decades, TimeUnit.Centuries); // 2
 * ```
 */
function convertTime(value, unit, targetUnit) {
    return (value * nanoConversions[unit]) / nanoConversions[targetUnit];
}
/**
 * Constructs a string representation of an elapsed amount of time. The least significant unit will be rounded to the nearest whole number.
 *
 * ```ts
 * elapsedToString(2001, ["s", "ms"]); // "2 s 1 ms"
 * elapsedToString(15, ["wks", "d"], ( elapsedUnit: TimeUnit.Days }); // "2 wks 1 d"
 * elapsedToString(1, [TimeUnit.Microseconds]); // "1000 μs"
 * ```
 *
 * @param showLeadingZeros whether to include the most significant units in the string if they are zero.
 */
function elapsedToString(elapsed, units, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.elapsedUnit, elapsedUnit = _c === void 0 ? TimeUnit.Milliseconds : _c, _d = _b.showLeadingZeros, showLeadingZeros = _d === void 0 ? true : _d;
    elapsed = roundToMultipleOf(convertTime(1, last(units), elapsedUnit), elapsed);
    var showZeros = showLeadingZeros;
    var tokens = [];
    units.forEach(function (unit, i) {
        var conversion = convertTime(1, elapsedUnit, unit);
        var value = Math.floor(elapsed * conversion);
        if (value > 0 || showZeros || i === units.length - 1) {
            tokens.push(value, unit);
            showZeros = true;
        }
        elapsed -= value / conversion;
    });
    return tokens.join(" ");
}
/** @hidden */
function getNanoConversions(unit, singular, nanos, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.aliases, aliases = _c === void 0 ? [] : _c, _d = _b.plural, plural = _d === void 0 ? singular + "s" : _d;
    return mapAsKeys(__spread([
        unit,
        singular,
        singular.toLowerCase(),
        plural,
        plural.toLowerCase()
    ], aliases), function () { return nanos; });
}

/**
 * Throws an error if `condition` is falsy, and acts as a type guard.
 *
 * @param message The message to set in the error that is thrown
 *
 * ```ts
 * class ProblemLogger {
 *   problem?: string;
 *
 *   log() {
 *     assert(this.problem, "You must set problem before logging it");
 *
 *     // now typescript knows `this.problem` is truthy, so we can safely call methods on it
 *     console.log(problem.toUpperCase());
 *   }
 * }
 * ```
 */
function assert(condition, message) {
    if (!condition) {
        throw new Error(message);
    }
}

/**
 * Extend this for classes whose objects are directly callable.
 *
 * ```ts
 * class Multiplier extends CallableObject<(value: number) => number> {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 */
var CallableObject = /** @class */ (function () {
    function CallableObject(callAction) {
        var _newTarget = this.constructor;
        return Object.setPrototypeOf(callAction, _newTarget.prototype);
    }
    return CallableObject;
}());

/** @hidden */
var seq = 0;
/**
 * Creates a factory function for building objects of a given type. Commonly used to create test objects for use in specs.
 *
 * ```ts
 * interface Message {
 *   id: number;
 *   text: string;
 * }
 *
 * interface Options {
 *   cypherDistance: number;
 * }
 *
 * declare function shiftCharacters(text: string, distance: number): string;
 *
 * const buildMessage = createBuilder<Message, Options>(
 *   (seq) => ({ id: seq, text: `message ${seq}` }),
 *   (message, _seq, options) => {
 *     message.text = shiftCharacters(message.text, options.cypherDistance || 0);
 *   },
 * );
 *
 * buildMessage(); // { id: 1, text: "message 1" }
 * buildMessage({ text: "Hello world!" }); // { id: 2, text: "Hello world!" }
 * buildMessage({ text: "abc" }, { cypherDistance: 3 }); // { id: 3, text: "def" }
 * ```
 */
function createBuilder(buildDefaults, afterBuild) {
    return function (attributes, options) {
        if (options === void 0) { options = {}; }
        ++seq;
        var obj = Object.assign(buildDefaults(seq, options), attributes);
        if (afterBuild) {
            afterBuild(obj, seq, options);
        }
        return obj;
    };
}

/**
 * Checks if `value` is anything other than `undefined`.
 *
 * ```ts
 * isDefined({ a: 1 }); // true
 * isDefined(false); // true
 * isDefined(null); // true
 * isDefined(undefined); // false
 * ```
 */
function isDefined(value) {
    return !isUndefined(value);
}

/**
 * Performs a deep comparison between two values to determine if they are equivalent, up to the given depth. Once that depth is reached, remaining values are compared using `Object.is()`.
 *
 * ```ts
 * let object1 = "a";
 * let object2 = "a";
 * isEqualAtDepth(0, object1, object2); // true
 * isEqualAtDepth(1, object1, object2); // true
 *
 * let object1 = { a: 1 };
 * let object2 = { a: 1 };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // true
 *
 * object1 = { a: [1, 2, 3], d: { e: 1 } };
 * object2 = { a: [1, 2, 3], d: { e: 1 } };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // false
 * isEqualAtDepth(2, object1, object2); // true
 * isEqualAtDepth(3, object1, object2); // true
 * ```
 */
function isEqualAtDepth(depth, value, other) {
    if (Object.is(value, other)) {
        // covers e.g. NaN === NaN
        return true;
    }
    if (depth === 0 || !(value instanceof Object && other instanceof Object)) {
        return false;
    }
    return hasSameValues(depth, value, other);
}
/** @hidden */
function hasSameValues(depth, value, other) {
    var e_1, _a, e_2, _b;
    try {
        for (var _c = __values(keys(value)), _d = _c.next(); !_d.done; _d = _c.next()) {
            var key = _d.value;
            if (!other.hasOwnProperty(key)) {
                return false;
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_d && !_d.done && (_a = _c.return)) _a.call(_c);
        }
        finally { if (e_1) throw e_1.error; }
    }
    try {
        for (var _e = __values(keys(other)), _f = _e.next(); !_f.done; _f = _e.next()) {
            var key = _f.value;
            if (!isEqualAtDepth(depth - 1, value[key], other[key])) {
                return false;
            }
        }
    }
    catch (e_2_1) { e_2 = { error: e_2_1 }; }
    finally {
        try {
            if (_f && !_f.done && (_b = _e.return)) _b.call(_e);
        }
        finally { if (e_2) throw e_2.error; }
    }
    return true;
}

/**
 * Converts a 2D array to a csv string. Values are converted using micro-dash's toString().
 *
 * ```ts
 * toCsv([["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"]]); // "a,b,c\nd,e,f\ng,h,i"
 * toCsv([
 *   ["a", "", "string"]
 *   [undefined, null],
 *   [true, false],
 *   [1, 2, 3],
 *   [{}, { hi: "there" }]
 * ]) // "a,,string\n,\ntrue,false\n1,2,3\n[object Object],[object Object]"
 * ```
 */
function toCsv(content) {
    return content
        .map(function (row) { return row.map(function (cell) { return toCellString(cell); }).join(","); })
        .join("\n");
}
/** @hidden */
var specialCsvCharactersRegexp = /["|,|\n|\r]/;
/** @hidden */
var allDoubleQuotes = /"/g;
/** @hidden */
function toCellString(value) {
    var str = toString(value);
    if (specialCsvCharactersRegexp.test(str)) {
        return "\"" + str.replace(allDoubleQuotes, "\"\"") + "\"";
    }
    else {
        return str;
    }
}

/**
 * Replaces a method on `object` with a wrapped version that will call the provided hooks in addition to the original method. See `wrapFunction()` for more details on the hooks.
 *
 * @returns a function to reset the method to its previous, unwrapped state
 *
 * ```ts
 * // log all get requests to the console
 * wrapMethod(HttpClient.prototype, "get", {
 *   before(url) {
 *     console.log("Sending GET request to", url);
 *   }
 * });
 *
 * // suppress benign error messages
 * const unwrap = wrapMethod(console, "error", {
 *   around(original, ...args) {
 *     if (args[0].message !== 'something benign') {
 *       original(...args);
 *     }
 *   }
 * });
 *
 * // remove error suppression (from above)
 * unwrap();
 * ```
 */
function wrapMethod(object, key, hooks) {
    var original = object[key];
    object[key] = wrapFunction(original, hooks);
    return function () {
        object[key] = original;
    };
}

/*
 * Public API Surface of s-js-utils
 */

/**
 * Generated bundle index. Do not edit.
 */

export { CallableObject, Debouncer, Deferred, TimeUnit, assert, convertTime, createBuilder, elapsedToString, isDefined, isEqualAtDepth, isSetEqual, isSuperset, mapAsKeys, mapToObject, roundToMultipleOf, setDifference, setIntersection, setUnion, sleep, symmetricSetDifference, toCsv, wrapFunction, wrapMethod };
//# sourceMappingURL=s-js-utils.js.map
